
// simple hybrid model simulation example (no state events)
// use "gcc -DDEBUG" for calendar dumps
// use "gcc -lm"

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

//////////////////////////////////////////////////////////////////
void error(const char *s)
{
    fprintf(stderr, "%s\n", s);
    exit(1);
}

//////////////////////////////////////////////////////////////////
typedef void (*ev_t)(void);     // pointer to event description

struct ev_notice {              // calendar item
    double atime;               // activation time
    int priority;               // for sorting of the same time events
    ev_t evptr;                 // event function pointer
};
void cal_dump_evn(struct ev_notice e);  // dump record (for debugging)

//////////////////////////////////////////////////////////////////
#define CAL_MAX 100             // limit of calendar size
struct ev_notice calendar_data[CAL_MAX];
unsigned calendar_size = 0;

//////////////////////////////////////////////////////////////////
void cal_init(void)             // initialize calendar
{
    calendar_size = 0;
}

struct ev_notice cal_get_first(void)    // return first entry
{
    if (calendar_size == 0)
        error("empty calendar");
    return calendar_data[0];
}

void cal_pop(void)              // remove first entry
{
    if (calendar_size == 0)
        error("empty calendar");
    memmove(calendar_data, calendar_data + 1,
            sizeof(struct ev_notice) * (calendar_size - 1));
    calendar_size -= 1;
}

void cal_insert(ev_t ptr, int prio, double t)   // add record
{
    unsigned oldsize = calendar_size;
    if (oldsize + 1 > (sizeof(calendar_data) / sizeof(calendar_data[0])))
        error("calendar overflow");
    // search (sort order: time,priority,FIFO)
    int i = 0;
    while (i < oldsize && calendar_data[i].atime < t)
        ++i;
    while (i < oldsize && calendar_data[i].atime == t
           && calendar_data[i].priority >= prio)
        ++i;
    // create space for new record
    if (i < oldsize) {
        memmove(calendar_data + i + 1, calendar_data + i,
                sizeof(struct ev_notice) * (oldsize - i));
    }
    calendar_size = oldsize + 1;
    calendar_data[i].atime = t;
    calendar_data[i].priority = prio;
    calendar_data[i].evptr = ptr;

#ifdef DEBUG
    printf("# Scheduling calendar[%d/%u] ", i, calendar_size);
    cal_dump_evn(calendar_data[i]);
    printf("\n");
#endif
}

//////////////////////////////////////////////////////////////////
// END event simplifies end condition testing, etc.
void END(void)
{
}

//////////////////////////////////////////////////////////////////
void cal_dump_evn(struct ev_notice e)   // dump record (for debugging)
{
    printf("(");
    printf("t=%g,", e.atime);
    printf("p=%d,", e.priority);
    if (e.evptr == END)
        printf("e=END");
    else
        printf("e=%p", e.evptr);
    printf(")");
}

//////////////////////////////////////////////////////////////////
void cal_dump(void)             // dump all records (for debugging)
{
#ifdef DEBUG
    printf("# calendar {\n");
    for (int i = 0; i < calendar_size; ++i) {
        printf("#  %2d: ", i);
        cal_dump_evn(calendar_data[i]);
        printf("\n");
    }
    printf("# }\n");
#endif
}

//////////////////////////////////////////////////////////////////
// model time
double time = 0;

//////////////////////////////////////////////////////////////////
// continuous model
const int size = 2;
double y[2] = { 0, 1 };         // initial conditions

void Dynamic(double y[], unsigned size, double in[])
{
    // size can be used for range-checking
    if (size != 2)
        error("unexpected state vector size");
    in[0] = y[1];               // y'
    in[1] = -y[0];              // y'' = -y
}

//////////////////////////////////////////////////////////////////
// discrete model
void event1(void)
{
    static int count = 0;
    printf("Time:%4g\t", time);
    printf("event1_%d\t", ++count);
    printf("y=% g \t err=% .3g\n", y[0], y[0]-sin(time));
    cal_insert(event1, 0, time + 1);
}

void event2(void)
{
    static int count = 0;
    printf("Time:%4g\t", time);
    printf("event2_%d\n", ++count);
}

//////////////////////////////////////////////////////////////////
// simple numerical integration method implementation
void euler_step(double step)
{
    double in[size];            // integrator inputs
    Dynamic(y, size, in);       // continuous model
    for (int i = 0; i < size; ++i)
        y[i] += step * in[i];   // Euler step
    time += step;               // advance time
}

//////////////////////////////////////////////////////////////////
// experiment control
int main()
{
    // initialization of model+simulator
    time = 0;
    double step = 0.001;          // try 1.0/16

    cal_init();

    cal_insert(END, -100, 1e2);   // set simulation end time
    cal_insert(event1, 0, time);
    cal_insert(event2, 0, 5);
    cal_insert(event2, 0, 5);
    cal_insert(event2, 1, 5);   // Priority: will be first event2 executed

    cal_dump();

    // start of simulation run
    struct ev_notice e;
    while ((e = cal_get_first()).evptr != END) {
        cal_pop();
        // continuous simulation to next event activation time
        while (time < e.atime) {
            if (time + step * 1.01 > e.atime) {
                // printf("### Time:%g  \treduced_step=%.16g\n", time, e.atime - time);
                euler_step(e.atime - time);
            } else
                euler_step(step);
        }
        //time = e.atime;         // no need for this
        e.evptr();              // event
    }
    time = e.atime;             // END time
    // end of simulation run

    cal_dump();
    return 0;
}

